#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2020 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Valerio Cosentino <valcos@bitergia.com>
#     inishchith <inishchith@gmail.com>
#

import warnings

import lizard
from graal.backends.core.analyzers.cloc import Cloc
from .analyzer import Analyzer


class Lizard(Analyzer):
    """A wrapper for Lizard, a code complexity analyzer, which is able
    to handle many imperative programming languages such as:
        C/C++ (works with C++14)
        Java
        C# (C Sharp)
        JavaScript
        Objective C
        Swift
        Python
        Ruby
        TTCN-3
        PHP
        Scala
        GDScript
        Golang
        Lua
    """
    version = '0.3.1'

    def __analyze_file(self, file_path, details):
        """Add code complexity information for a file using Lizard.

        Current information includes cyclomatic complexity (ccn),
        avg lines of code and tokens, number of functions and tokens.
        Optionally, the following information can be included for every function:
        ccn, tokens, LOC, lines, name, args, start, end

        :param file_path: file path
        :param details: if True, it returns information about single functions

        :returns  result: dict of the results of the analysis
        """
        result = {}

        # Filter DeprecationWarning from lizard_ext/auto_open.py line 26
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=DeprecationWarning)
            analysis = lizard.analyze_file(file_path)

        result['ccn'] = analysis.CCN
        result['avg_ccn'] = analysis.average_cyclomatic_complexity
        result['avg_loc'] = analysis.average_nloc
        result['avg_tokens'] = analysis.average_token_count
        result['num_funs'] = len(analysis.function_list)
        result['loc'] = analysis.nloc
        result['tokens'] = analysis.token_count
        result['ext'] = file_path.split(".")[-1]

        if not details:
            return result

        funs_data = []
        for fun in analysis.function_list:
            fun_data = {'ccn': fun.cyclomatic_complexity,
                        'tokens': fun.token_count,
                        'loc': fun.nloc,
                        'lines': fun.length,
                        'name': fun.name,
                        'args': fun.parameter_count,
                        'start': fun.start_line,
                        'end': fun.end_line}
            funs_data.append(fun_data)

        result['funs'] = funs_data
        return result

    def __analyze_repository(self, repository_path, files_affected, details):
        """Add code complexity information for a given repository
        using Lizard and CLOC.

        Current information includes cyclomatic complexity (ccn),
        lines of code, number of functions, tokens, blanks and comments.

        :param repository_path: repository path
        :param details: if True, it returns fine-grained results

        :returns  result: list of the results of the analysis
        """
        analysis_result = []

        repository_analysis = lizard.analyze(
            paths=[repository_path],
            threads=1,
            exts=lizard.get_extensions([]),
        )
        cloc = Cloc()

        for analysis in repository_analysis:
            cloc_analysis = cloc.analyze(file_path=analysis.filename)
            file_path = analysis.filename.replace(repository_path + "/", '')
            in_commit = True if file_path in files_affected else False

            result = {
                'loc': analysis.nloc,
                'ccn': analysis.CCN,
                'tokens': analysis.token_count,
                'num_funs': len(analysis.function_list),
                'file_path': file_path,
                'in_commit': in_commit,
                'blanks': cloc_analysis['blanks'],
                'comments': cloc_analysis['comments']
            }
            analysis_result.append(result)

        # TODO: implement details option

        return analysis_result

    def analyze(self, **kwargs):
        """Add code complexity information using Lizard.

        :param file_path: file path
        :param repository_path: repository path
        :param details: if True, it returns detailed information about an analysis

        :returns  result: the results of the analysis
        """

        details = kwargs['details']

        if kwargs.get('repository_level', False):
            files_affected = kwargs['files_affected']
            result = self.__analyze_repository(kwargs["repository_path"], files_affected, details)
        else:
            result = self.__analyze_file(kwargs['file_path'], details)

        return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2020 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Valerio Cosentino <valcos@bitergia.com>
#     inishchith <inishchith@gmail.com>
#

import os
import unittest.mock

from graal.graal import GraalError
from graal.graal import GraalCommandArgumentParser
from graal.backends.core.analyzers.cloc import Cloc
from graal.backends.core.analyzers.lizard import Lizard
from graal.backends.core.cocom import (CATEGORY_COCOM_LIZARD_FILE,
                                       CATEGORY_COCOM_LIZARD_REPOSITORY,
                                       CATEGORY_COCOM_SCC_FILE,
                                       CATEGORY_COCOM_SCC_REPOSITORY,
                                       CoCom,
                                       FileAnalyzer,
                                       RepositoryAnalyzer,
                                       CoComCommand)
from perceval.utils import DEFAULT_DATETIME
from base_analyzer import (ANALYZER_TEST_FILE,
                           TestCaseAnalyzer)
from base_repo import TestCaseRepo


class TestCoComBackend(TestCaseRepo):
    """CoCom backend tests"""

    def test_initialization(self):
        """Test whether attributes are initializated"""

        cc = CoCom('http://example.com', self.git_path, self.worktree_path, tag='test')
        self.assertEqual(cc.uri, 'http://example.com')
        self.assertEqual(cc.gitpath, self.git_path)
        self.assertEqual(cc.worktreepath, os.path.join(self.worktree_path, os.path.split(cc.gitpath)[1]))
        self.assertEqual(cc.origin, 'http://example.com')
        self.assertEqual(cc.tag, 'test')

        cc = CoCom('http://example.com', self.git_path, self.worktree_path, details=True, tag='test')
        self.assertEqual(cc.uri, 'http://example.com')
        self.assertEqual(cc.gitpath, self.git_path)
        self.assertEqual(cc.worktreepath, os.path.join(self.worktree_path, os.path.split(cc.gitpath)[1]))
        self.assertEqual(cc.origin, 'http://example.com')
        self.assertEqual(cc.tag, 'test')

        # When tag is empty or None it will be set to the value in uri
        cc = CoCom('http://example.com', self.git_path, self.worktree_path)
        self.assertEqual(cc.origin, 'http://example.com')
        self.assertEqual(cc.tag, 'http://example.com')

        cc = CoCom('http://example.com', self.git_path, self.worktree_path)
        self.assertEqual(cc.origin, 'http://example.com')
        self.assertEqual(cc.tag, 'http://example.com')

    def test_fetch_lizard_file(self):
        """Test whether commits are properly processed via file level"""

        cc = CoCom('http://example.com', self.git_path, self.worktree_path, in_paths=['perceval/backends/core/github.py'])
        commits = [commit for commit in cc.fetch()]

        self.assertEqual(len(commits), 1)
        self.assertFalse(os.path.exists(cc.worktreepath))

        for commit in commits:
            self.assertEqual(commit['backend_name'], 'CoCom')
            self.assertEqual(commit['category'], CATEGORY_COCOM_LIZARD_FILE)
            self.assertEqual(commit['data']['analysis'][0]['file_path'],
                             'perceval/backends/core/github.py')
            self.assertTrue('Author' in commit['data'])
            self.assertTrue('Commit' in commit['data'])
            self.assertTrue('files' in commit['data'])
            self.assertTrue('parents' in commit['data'])
            self.assertFalse('refs' in commit['data'])

    def test_fetch_scc_file(self):
        """Test whether commits are properly processed via file level"""

        cc = CoCom('http://example.com', self.git_path, self.worktree_path, in_paths=['perceval/backends/core/github.py'])
        commits = [commit for commit in cc.fetch(category="code_complexity_scc_file")]

        self.assertEqual(len(commits), 1)
        self.assertFalse(os.path.exists(cc.worktreepath))

        for commit in commits:
            self.assertEqual(commit['backend_name'], 'CoCom')
            self.assertEqual(commit['category'], CATEGORY_COCOM_SCC_FILE)
            self.assertEqual(commit['data']['analysis'][0]['file_path'],
                             'perceval/backends/core/github.py')
            self.assertTrue('Author' in commit['data'])
            self.assertTrue('Commit' in commit['data'])
            self.assertTrue('files' in commit['data'])
            self.assertTrue('parents' in commit['data'])
            self.assertFalse('refs' in commit['data'])

    def test_fetch_lizard_repository(self):
        """Test whether commits are properly processed via repository level"""

        cc = CoCom('http://example.com', self.git_path, self.worktree_path)
        commits = [commit for commit in cc.fetch(category="code_complexity_lizard_repository")]

        self.assertEqual(len(commits), 6)
        self.assertFalse(os.path.exists(cc.worktreepath))

        for commit in commits:
            self.assertEqual(commit['backend_name'], 'CoCom')
            self.assertEqual(commit['category'], CATEGORY_COCOM_LIZARD_REPOSITORY)
            self.assertTrue('Author' in commit['data'])
            self.assertTrue('Commit' in commit['data'])
            self.assertTrue('files' in commit['data'])
            self.assertTrue('parents' in commit['data'])
            self.assertFalse('refs' in commit['data'])

    def test_fetch_scc_repository(self):
        """Test whether commits are properly processed via repository level"""

        cc = CoCom('http://example.com', self.git_path, self.worktree_path)
        commits = [commit for commit in cc.fetch(category="code_complexity_scc_repository")]

        self.assertEqual(len(commits), 6)
        self.assertFalse(os.path.exists(cc.worktreepath))

        for commit in commits:
            self.assertEqual(commit['backend_name'], 'CoCom')
            self.assertEqual(commit['category'], CATEGORY_COCOM_SCC_REPOSITORY)
            self.assertTrue('Author' in commit['data'])
            self.assertTrue('Commit' in commit['data'])
            self.assertTrue('files' in commit['data'])
            self.assertTrue('parents' in commit['data'])
            self.assertFalse('refs' in commit['data'])

    def test_fetch_unknown(self):
        """Test whether commits are properly processed"""

        cc = CoCom('http://example.com', self.git_path, self.worktree_path)

        with self.assertRaises(GraalError):
            _ = cc.fetch(category="unknown")

    def test_fetch_analysis(self):
        """Test whether commits have properly set values"""

        cc = CoCom('http://example.com', self.git_path, self.worktree_path, details=True)
        commits = [commit for commit in cc.fetch()]

        self.assertEqual(len(commits), 6)
        self.assertFalse(os.path.exists(cc.worktreepath))

        deleted_file_commit = commits[5]

        self.assertEqual(deleted_file_commit['data']['analysis'][0]['file_path'],
                         'perceval/backends/graal.py')
        self.assertEqual(deleted_file_commit['data']['analysis'][0]['blanks'], None)
        self.assertEqual(deleted_file_commit['data']['analysis'][0]['comments'], None)
        self.assertEqual(deleted_file_commit['data']['analysis'][0]['loc'], None)
        self.assertEqual(deleted_file_commit['data']['analysis'][0]['ccn'], None)
        self.assertEqual(deleted_file_commit['data']['analysis'][0]['avg_ccn'], None)
        self.assertEqual(deleted_file_commit['data']['analysis'][0]['avg_loc'], None)
        self.assertEqual(deleted_file_commit['data']['analysis'][0]['avg_tokens'], None)
        self.assertEqual(deleted_file_commit['data']['analysis'][0]['num_funs'], None)
        self.assertEqual(deleted_file_commit['data']['analysis'][0]['tokens'], None)
        self.assertEqual(deleted_file_commit['data']['analysis'][0]['funs'], [])

    def test_metadata_category(self):
        """Test metadata_category"""
        item = {
            "Author": "Nishchith Shetty <inishchith@gmail.com>",
            "AuthorDate": "Tue Feb 26 22:06:31 2019 +0530",
            "Commit": "Nishchith Shetty <inishchith@gmail.com>",
            "CommitDate": "Tue Feb 26 22:06:31 2019 +0530",
            "analysis": [],
            "analyzer": "lizard_file",
            "commit": "5866a479587e8b548b0cb2d591f3a3f5dab04443",
            "message": "[copyright] Update copyright dates"
        }
        self.assertEqual(CoCom.metadata_category(item), CATEGORY_COCOM_LIZARD_FILE)

        item = {
            "Author": "Nishchith Shetty <inishchith@gmail.com>",
            "AuthorDate": "Tue Feb 26 22:06:31 2019 +0530",
            "Commit": "Nishchith Shetty <inishchith@gmail.com>",
            "CommitDate": "Tue Feb 26 22:06:31 2019 +0530",
            "analysis": [],
            "analyzer": "lizard_repository",
            "commit": "5866a479587e8b548b0cb2d591f3a3f5dab04443",
            "message": "[copyright] Update copyright dates"
        }
        self.assertEqual(CoCom.metadata_category(item), CATEGORY_COCOM_LIZARD_REPOSITORY)

        item = {
            "Author": "Nishchith Shetty <inishchith@gmail.com>",
            "AuthorDate": "Tue Feb 26 22:06:31 2019 +0530",
            "Commit": "Nishchith Shetty <inishchith@gmail.com>",
            "CommitDate": "Tue Feb 26 22:06:31 2019 +0530",
            "analysis": [],
            "analyzer": "unknown",
            "commit": "5866a479587e8b548b0cb2d591f3a3f5dab04443",
            "message": "[copyright] Update copyright dates"
        }
        with self.assertRaises(GraalError):
            _ = CoCom.metadata_category(item)


class TestFileAnalyzer(TestCaseAnalyzer):
    """FileAnalyzer tests"""

    def test_init(self):
        """Test initialization"""

        file_analyzer = FileAnalyzer()

        self.assertIsInstance(file_analyzer, FileAnalyzer)
        self.assertIsInstance(file_analyzer.cloc, Cloc)
        self.assertIsInstance(file_analyzer.lizard, Lizard)
        self.assertFalse(file_analyzer.details)

        file_analyzer = FileAnalyzer(details=True)

        self.assertIsInstance(file_analyzer, FileAnalyzer)
        self.assertIsInstance(file_analyzer.cloc, Cloc)
        self.assertIsInstance(file_analyzer.lizard, Lizard)
        self.assertTrue(file_analyzer.details)

    def test_analyze_no_functions(self):
        """Test whether the analyze method works"""

        file_path = os.path.join(self.tmp_data_path, ANALYZER_TEST_FILE)
        file_analyzer = FileAnalyzer()
        analysis = file_analyzer.analyze(file_path)

        self.assertNotIn('funs', analysis)
        self.assertIn('ccn', analysis)
        self.assertIn('avg_loc', analysis)
        self.assertIn('avg_tokens', analysis)
        self.assertIn('loc', analysis)
        self.assertIn('tokens', analysis)
        self.assertIn('blanks', analysis)
        self.assertIn('comments', analysis)

    def test_analyze_functions(self):
        """Test whether the analyze method returns functions information"""

        file_path = os.path.join(self.tmp_data_path, ANALYZER_TEST_FILE)
        file_analyzer = FileAnalyzer(details=True)
        analysis = file_analyzer.analyze(file_path)

        self.assertIn('ccn', analysis)
        self.assertIn('avg_loc', analysis)
        self.assertIn('avg_tokens', analysis)
        self.assertIn('loc', analysis)
        self.assertIn('tokens', analysis)
        self.assertIn('blanks', analysis)
        self.assertIn('comments', analysis)
        self.assertIn('funs', analysis)

        for fd in analysis['funs']:
            self.assertIn('ccn', fd)
            self.assertIn('tokens', fd)
            self.assertIn('loc', fd)
            self.assertIn('lines', fd)
            self.assertIn('name', fd)
            self.assertIn('args', fd)
            self.assertIn('start', fd)
            self.assertIn('end', fd)


class TestRepositoryAnalyzer(TestCaseAnalyzer):
    """RepositoryAnalyzer tests"""

    def test_init(self):
        """Test initialization"""

        repository_analyzer = RepositoryAnalyzer()

        self.assertIsInstance(repository_analyzer, RepositoryAnalyzer)
        self.assertIsInstance(repository_analyzer.analyzer, Lizard)
        self.assertFalse(repository_analyzer.details)

        repository_analyzer = RepositoryAnalyzer(details=True)

        self.assertIsInstance(repository_analyzer, RepositoryAnalyzer)
        self.assertIsInstance(repository_analyzer.analyzer, Lizard)
        self.assertTrue(repository_analyzer.details)

    def test_analyze(self):
        """Test whether the analyze method works"""

        repository_path = self.tmp_data_path
        repository_analyzer = RepositoryAnalyzer()
        analysis = repository_analyzer.analyze(repository_path, files_affected=[])

        file_analysis = analysis[0]

        self.assertIn('num_funs', file_analysis)
        self.assertIn('ccn', file_analysis)
        self.assertIn('loc', file_analysis)
        self.assertIn('tokens', file_analysis)
        self.assertIn('file_path', file_analysis)
        self.assertIn('in_commit', file_analysis)
        self.assertIn('blanks', file_analysis)
        self.assertIn('comments', file_analysis)


class TestCoComCommand(unittest.TestCase):
    """CoComCommand tests"""

    def test_backend_class(self):
        """Test if the backend class is CoCom"""

        self.assertIs(CoComCommand.BACKEND, CoCom)

    def test_setup_cmd_parser(self):
        """Test setup_cmd_parser"""

        parser = CoComCommand.setup_cmd_parser()
        self.assertIsInstance(parser, GraalCommandArgumentParser)
        self.assertEqual(parser._backend, CoCom)

        args = ['http://example.com/',
                '--git-path', '/tmp/gitpath',
                '--tag', 'test',
                '--from-date', '1970-01-01']

        parsed_args = parser.parse(*args)
        self.assertEqual(parsed_args.uri, 'http://example.com/')
        self.assertEqual(parsed_args.git_path, '/tmp/gitpath')
        self.assertEqual(parsed_args.tag, 'test')
        self.assertEqual(parsed_args.from_date, DEFAULT_DATETIME)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2020 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Valerio Cosentino <valcos@bitergia.com>
#     inishchith <inishchith@gmail.com>
#

import io
import os
import shutil
import tarfile
import tempfile
import unittest
import unittest.mock

from grimoirelab_toolkit.datetime import str_to_datetime
from perceval.errors import RepositoryError
from perceval.utils import DEFAULT_DATETIME, DEFAULT_LAST_DATETIME

import graal
from graal.graal import (DEFAULT_WORKTREE_PATH,
                         CATEGORY_GRAAL,
                         GIT_EXEC_PATH,
                         Graal,
                         GraalCommand,
                         GraalRepository,
                         GraalCommandArgumentParser,
                         logger)
from base_repo import TestCaseRepo


CATEGORY_MOCKED = 'mocked'


class MockedGraalRepository(GraalRepository):

    RAISE_EXCEPTION = False

    def __init__(self, uri, dirpath, raise_exception):
        super().__init__(uri, dirpath)
        self.worktreepath = None
        MockedGraalRepository.RAISE_EXCEPTION = raise_exception

    @staticmethod
    def _exec(cmd, cwd=None, env=None, ignored_error_codes=None,
              encoding='utf-8'):

        if MockedGraalRepository.RAISE_EXCEPTION:
            if cmd[:2] == [GIT_EXEC_PATH, "archive"]:
                raise OSError

            raise RepositoryError(cause='oops!')

        super()._exec(cmd, cwd=cwd, env=env, ignored_error_codes=ignored_error_codes,
                      encoding=encoding)


class MockedGraal(Graal):

    CATEGORIES = [CATEGORY_MOCKED]

    def __init__(self, uri, gitpath, worktreepath=DEFAULT_WORKTREE_PATH,
                 entrypoint=None, in_paths=None, out_paths=None, details=False,
                 tag=None, archive=None, raise_exception=False):
        super().__init__(uri, gitpath, worktreepath=worktreepath, entrypoint=entrypoint,
                         in_paths=in_paths, out_paths=out_paths, details=details,
                         tag=tag, archive=archive)
        self.raise_exception = raise_exception

    def fetch(self, category=CATEGORY_MOCKED, paths=None,
              from_date=DEFAULT_DATETIME, to_date=DEFAULT_LAST_DATETIME,
              branches=None, latest_items=False):
        """Fetch commits and add code complexity information."""

        items = super().fetch(category,
                              from_date=from_date, to_date=to_date,
                              branches=branches, latest_items=latest_items)

        return items

    @staticmethod
    def metadata_category(item):
        """Extracts the category from a MockedGraal item.

        This backend only generates one type of item which is
        'mocked'.
        """
        return CATEGORY_MOCKED

    def _filter_commit(self, commit):
        """Filter a commit according to its data (e.g., author, sha, etc.)

        :param commit: a Perceval commit item

        :returns: a boolean value
        """
        return False

    def _analyze(self, commit, paths=None):
        """Analyse a commit and the corresponding
        checkout version of the repository

        :param commit: a Perceval commit item
        :param paths: a list of paths to narrow the analysis
        """
        mods = 0
        for f in commit['files']:
            added = int(f['added']) if 'added' in f else 0
            removed = int(f['removed']) if 'removed' in f else 0
            mods += (added - removed)

        return {'lines_modified': mods}

    def _post(self, commit):
        """Perform operation (e.g., removing attributes) on the Graal item obtained

        :param commit: a Graal commit item
        """
        if self.raise_exception:
            raise Exception

        commit.pop('Author', None)
        commit.pop('Commit', None)
        commit.pop('files', None)
        commit.pop('parents', None)
        commit.pop('refs', None)
        return commit


class MockedGraalCommand(GraalCommand):
    BACKEND = MockedGraal

    @classmethod
    def setup_cmd_parser(cls):

        parser = GraalCommand.setup_cmd_parser(cls.BACKEND)

        return parser


class CommandBackend(MockedGraal):
    """Backend used for testing in BackendCommand tests"""

    def fetch_items(self, category, **kwargs):
        for item in super().fetch_items(category, **kwargs):
            yield item


class ErrorCommandBackend(CommandBackend):
    """Backend which raises an exception while fetching items"""

    def fetch_items(self, category, **kwargs):
        for item in super().fetch_items(category, **kwargs):
            yield item
            raise Exception


class TestGraalBackend(TestCaseRepo):
    """Graal backend tests"""

    def test_initialization(self):
        """Test whether attributes are initializated"""

        graal = Graal('http://example.com', self.git_path, self.worktree_path, tag='test')
        self.assertEqual(graal.uri, 'http://example.com')
        self.assertEqual(graal.gitpath, self.git_path)
        self.assertEqual(graal.worktreepath, os.path.join(self.worktree_path, os.path.split(graal.gitpath)[1]))
        self.assertEqual(graal.origin, 'http://example.com')
        self.assertEqual(graal.tag, 'test')
        self.assertIsNone(graal.entrypoint)
        self.assertIsNone(graal.in_paths)
        self.assertIsNone(graal.out_paths)
        self.assertFalse(graal.details)
        self.assertIsNone(graal.exec_path)

        # When tag is empty or None it will be set to the value in uri
        graal = Graal('http://example.com', self.git_path, self.worktree_path)
        self.assertEqual(graal.origin, 'http://example.com')
        self.assertEqual(graal.tag, 'http://example.com')
        self.assertIsNone(graal.exec_path)

        graal = Graal('http://example.com', self.git_path, self.worktree_path)
        self.assertEqual(graal.origin, 'http://example.com')
        self.assertEqual(graal.tag, 'http://example.com')
        self.assertIsNone(graal.exec_path)

        graal = Graal('http://example.com', self.git_path, self.worktree_path, exec_path='/tmp/nomossa')
        self.assertEqual(graal.origin, 'http://example.com')
        self.assertEqual(graal.tag, 'http://example.com')
        self.assertEqual(graal.exec_path, '/tmp/nomossa')

        graal = Graal('http://example.com', self.git_path,
                      entrypoint="entrypoint", in_paths=["x"], out_paths=["y"], details=True)
        self.assertEqual(graal.worktreepath, os.path.join(DEFAULT_WORKTREE_PATH, os.path.split(graal.gitpath)[1]))
        self.assertEqual(graal.entrypoint, "entrypoint")
        self.assertEqual(graal.in_paths, ["x"])
        self.assertEqual(graal.out_paths, ["y"])
        self.assertTrue(graal.details)
        self.assertIsNone(graal.exec_path)

    def test_fetch_no_analysis(self):
        """Test whether commits are inflated with the analysis attribute"""

        graal = Graal('http://example.com', self.git_path, self.worktree_path)
        commits = [commit for commit in graal.fetch()]

        self.assertEqual(len(commits), 6)
        self.assertFalse(os.path.exists(graal.worktreepath))

        for commit in commits:
            self.assertEqual(commit['backend_name'], 'Graal')
            self.assertEqual(commit['category'], CATEGORY_GRAAL)
            self.assertEqual(commit['data']['analysis'], {})

    def test_fetch_analysis(self):
        """Test whether commits are properly processed"""

        mocked = MockedGraal('http://example.com', self.git_path, self.worktree_path)
        commits = [commit for commit in mocked.fetch()]

        self.assertEqual(len(commits), 6)
        self.assertFalse(os.path.exists(mocked.worktreepath))

        commit = commits[0]
        self.assertEqual(commit['backend_name'], 'MockedGraal')
        self.assertEqual(commit['category'], CATEGORY_MOCKED)
        self.assertEqual(commit['data']['analysis']['lines_modified'], 4177)
        self.assertFalse('Author' in commit['data'])
        self.assertFalse('Commit' in commit['data'])
        self.assertFalse('files' in commit['data'])
        self.assertFalse('parents' in commit['data'])
        self.assertFalse('refs' in commit['data'])

        commit = commits[1]
        self.assertEqual(commit['backend_name'], 'MockedGraal')
        self.assertEqual(commit['category'], CATEGORY_MOCKED)
        self.assertEqual(commit['data']['analysis']['lines_modified'], 25)
        self.assertFalse('Author' in commit['data'])
        self.assertFalse('Commit' in commit['data'])
        self.assertFalse('files' in commit['data'])
        self.assertFalse('parents' in commit['data'])
        self.assertFalse('refs' in commit['data'])

        commit = commits[2]
        self.assertEqual(commit['backend_name'], 'MockedGraal')
        self.assertEqual(commit['category'], CATEGORY_MOCKED)
        self.assertEqual(commit['data']['analysis']['lines_modified'], 12)
        self.assertFalse('Author' in commit['data'])
        self.assertFalse('Commit' in commit['data'])
        self.assertFalse('files' in commit['data'])
        self.assertFalse('parents' in commit['data'])
        self.assertFalse('refs' in commit['data'])

    def test_fetch_analysis_on_error(self):
        mocked = MockedGraal('http://example.com', self.git_path, self.worktree_path, raise_exception=True)
        with self.assertRaises(Exception):
            _ = [commit for commit in mocked.fetch()]


class TestGraalRepository(TestCaseRepo):
    """GraalRepository tests"""

    def setUp(self):
        super().setUp()

        patcher = unittest.mock.patch('os.getenv')
        self.addCleanup(patcher.stop)
        self.mock_getenv = patcher.start()
        self.mock_getenv.return_value = ''

    def tearDown(self):
        repo = GraalRepository('http://example.git', self.git_path)
        try:
            repo._exec(['git', 'branch', '-D', 'testworktree'], cwd=repo.dirpath, env=repo.gitenv)
            repo._exec(['git', 'branch', '-D', 'graaltest'], cwd=repo.dirpath, env=repo.gitenv)
        except RepositoryError:
            pass

    def test_init(self):
        """Test initialization"""

        repo = GraalRepository('http://example.git', self.git_path)

        self.assertIsInstance(repo, GraalRepository)
        self.assertEqual(repo.uri, 'http://example.git')
        self.assertEqual(repo.dirpath, self.git_path)

        # Check command environment variables
        expected = {
            'LANG': 'C',
            'PAGER': '',
            'HOME': '',
            'HTTPS_PROXY': '',
            'HTTP_PROXY': '',
            'NO_PROXY': ''
        }
        self.assertDictEqual(repo.gitenv, expected)

    def test_worktree(self):
        """Test whether a working tree is created"""

        new_path = os.path.join(self.tmp_path, 'testworktree')

        repo = GraalRepository('http://example.git', self.git_path)
        self.assertIsNone(repo.worktreepath)
        repo.worktree(new_path)
        self.assertEqual(repo.worktreepath, new_path)
        self.assertTrue(os.path.exists(repo.worktreepath))

        repo.prune()
        self.assertFalse(os.path.exists(repo.worktreepath))

    def test_worktree_from_branch(self):
        """Test whether a working tree is created from a target branch"""

        new_path = os.path.join(self.tmp_path, 'testworktree')

        repo = GraalRepository('http://example.git', self.git_path)

        self.assertIsNone(repo.worktreepath)
        repo.worktree(new_path, branch='master')
        self.assertEqual(repo.worktreepath, new_path)
        self.assertTrue(os.path.exists(repo.worktreepath))

        repo.prune()
        self.assertFalse(os.path.exists(repo.worktreepath))

    def test_worktree_already_exists(self):
        """Test whether a debug info is logged when the worktree already exists"""

        new_path = os.path.join(self.tmp_path, 'testworktree')

        repo = GraalRepository('http://example.git', self.git_path)
        self.assertIsNone(repo.worktreepath)
        repo.worktree(new_path, branch='master')

        with self.assertLogs(logger, level='DEBUG') as cm:
            repo.worktree(new_path, branch='master')
            self.assertRegex(cm.output[0],
                             'DEBUG:graal.graal:Git worktree(?s:.)*not created(?s:.)*already exists(?s:.)*')

        repo.prune()
        self.assertFalse(os.path.exists(repo.worktreepath))

    def test_worktree_on_error(self):
        """Test whether a RepositoryError is thrown in case of error"""

        new_path = os.path.join(self.tmp_path, 'testworktree')

        repo = MockedGraalRepository('http://example.git', self.git_path, raise_exception=True)
        with self.assertRaises(RepositoryError):
            repo.worktree(new_path)

    def test_prune(self):
        """Test whether a working tree is deleted"""

        new_path = os.path.join(self.tmp_path, 'testworktree')

        repo = GraalRepository('http://example.git', self.git_path)
        repo.worktreepath = new_path
        repo.worktree(new_path)
        self.assertEqual(repo.worktreepath, new_path)
        self.assertTrue(os.path.exists(repo.worktreepath))

        repo.prune()
        self.assertFalse(os.path.exists(repo.worktreepath))

    def test_prune_on_error(self):
        """Test whether a RepositoryError is thrown in case of error"""

        repo = MockedGraalRepository('http://example.git', self.git_path, raise_exception=True)
        with self.assertRaises(RepositoryError):
            repo.prune()

    def test_checkout(self):
        """Test whether Git checkout commands are correctly executed"""

        new_path = os.path.join(self.tmp_path, 'testworktree')

        repo = GraalRepository('http://example.git', self.git_path)
        self.assertIsNone(repo.worktreepath)
        repo.worktree(new_path)

        repo.checkout("075f0c6161db5a3b1c8eca45e08b88469bb148b9")
        current_commit_hash = self.__git_show_hash(repo)
        self.assertEqual("075f0c6161db5a3b1c8eca45e08b88469bb148b9", current_commit_hash)

        repo.checkout("4f3b403d47fb291a9a942a62d62c24faa79244c8")
        current_commit_hash = self.__git_show_hash(repo)
        self.assertEqual("4f3b403d47fb291a9a942a62d62c24faa79244c8", current_commit_hash)

        repo.checkout("825b4da7ca740f7f2abbae1b3402908a44d130cd")
        current_commit_hash = self.__git_show_hash(repo)
        self.assertEqual("825b4da7ca740f7f2abbae1b3402908a44d130cd", current_commit_hash)

        repo.prune()
        self.assertFalse(os.path.exists(repo.worktreepath))

    def test_checkout_on_error(self):
        """Test whether a RepositoryError is thrown in case of error"""

        repo = MockedGraalRepository('http://example.git', self.git_path, raise_exception=True)
        with self.assertRaises(RepositoryError):
            repo.checkout("825b4da7ca740f7f2abbae1b3402908a44d130cd")

    def test_archive(self):
        """Test whether a Git archive command is correctly executed"""

        repo = GraalRepository('http://example.git', self.git_path)
        file_obj = repo.archive("825b4da7ca740f7f2abbae1b3402908a44d130cd")

        self.assertIsInstance(file_obj, io.BytesIO)

    def test_archive_on_error(self):
        """Test whether the method return nothing in case of error"""

        repo = MockedGraalRepository('http://example.git', self.git_path, raise_exception=True)
        file_obj = repo.archive("825b4da7ca740f7f2abbae1b3402908a44d130cd")
        self.assertIsNone(file_obj)

    def test_tar_obj(self):
        """Test whether a BytesIO object is converted to a tar object"""

        repo = GraalRepository('http://example.git', self.git_path)
        file_obj = repo.archive("825b4da7ca740f7f2abbae1b3402908a44d130cd")

        tar_obj = repo.tar_obj(file_obj)
        self.assertIsInstance(tar_obj, tarfile.TarFile)

    def test_tar_obj_on_error(self):
        """Test whether the method return nothing in case of error"""

        repo = GraalRepository('http://example.git', self.git_path)
        file_obj = io.BytesIO(b'abcdefghilmnopqrstuvz')
        tar_obj = repo.tar_obj(file_obj)
        self.assertIsNone(tar_obj)

    def test_filter_tar(self):
        """Test whether tar object members are filtered"""

        repo = GraalRepository('http://example.git', self.git_path)
        file_obj = repo.archive("825b4da7ca740f7f2abbae1b3402908a44d130cd")
        tar_obj = repo.tar_obj(file_obj)
        self.assertEqual(len(tar_obj.getmembers()), 18)

        to_select = ['.gitignore']
        filtered_obj = repo.filter_tar(tar_obj, to_select)
        self.assertEqual(len(filtered_obj.getmembers()), 1)
        self.assertEqual(filtered_obj.getmembers()[0].name, '.gitignore')

    def test_filter_tar_no_members(self):
        """Test whether an null object is returned if not members are present"""

        repo = GraalRepository('http://example.git', self.git_path)
        file_obj = repo.archive("825b4da7ca740f7f2abbae1b3402908a44d130cd")
        tar_obj = repo.tar_obj(file_obj)

        to_select = []
        filtered_obj = repo.filter_tar(tar_obj, to_select)
        self.assertIsNone(filtered_obj)

    def test_tar(self):
        """Test whether tar object is saved to disk"""

        tar_path = os.path.join(self.tmp_path, 'testtar.tar.gz')

        repo = GraalRepository('http://example.git', self.git_path)
        file_obj = repo.archive("825b4da7ca740f7f2abbae1b3402908a44d130cd")
        tar_obj = repo.tar_obj(file_obj)

        repo.tar(tar_obj, tar_path)

        self.assertTrue(os.path.exists(tar_path))
        self.assertTrue(os.path.isfile(tar_path))

        os.remove(tar_path)

    def test_exists(self):
        """Test whether the method exists works properly"""

        exists_path = os.path.join(self.tmp_path, 'existtest')

        self.assertFalse(GraalRepository.exists(exists_path))
        f = open(exists_path, 'w')
        f.close()
        self.assertTrue(GraalRepository.exists(exists_path))

    def test_untar(self):
        """Test whether tar file is untarred"""

        untar_path = os.path.join(self.tmp_path, 'untesttar')

        repo = GraalRepository('http://example.git', self.git_path)
        file_obj = repo.archive("825b4da7ca740f7f2abbae1b3402908a44d130cd")
        tar_obj = repo.tar_obj(file_obj)

        repo.untar(tar_obj, untar_path)
        self.assertTrue(os.path.exists(untar_path))
        self.assertTrue(os.path.isdir(untar_path))

        shutil.rmtree(untar_path)

    def test_extension(self):
        """Test whether file extensions are identified"""

        self.assertEqual(GraalRepository.extension('setup.py'), 'py')
        self.assertEqual(GraalRepository.extension('tests/requirements.txt'), 'txt')
        self.assertEqual(GraalRepository.extension('LICENSE'), 'LICENSE')

    def test_files(self):
        """Test whether all files in a directory and its sub-directories are shown"""

        new_path = os.path.join(self.tmp_path, 'testworktree')

        repo = GraalRepository('http://example.git', self.git_path)
        repo.worktree(new_path)

        expected = [os.path.join(new_path, 'perceval/archive.py'),
                    os.path.join(new_path, 'perceval/backend.py'),
                    os.path.join(new_path, 'perceval/client.py'),
                    os.path.join(new_path, 'perceval/__init__.py'),
                    os.path.join(new_path, 'perceval/_version.py'),
                    os.path.join(new_path, 'perceval/utils.py'),
                    os.path.join(new_path, 'perceval/errors.py'),
                    os.path.join(new_path, 'perceval/backends/__init__.py'),
                    os.path.join(new_path, 'perceval/backends/core/__init__.py'),
                    os.path.join(new_path, 'perceval/backends/core/github.py'),
                    os.path.join(new_path, 'perceval/backends/core/mbox.py'),
                    os.path.join(new_path, 'perceval/backends/core/git.py')
                    ]

        files = repo.files(new_path)

        self.assertEqual(len(files), len(expected))
        for f in files:
            self.assertIn(f, expected)

        repo.prune()
        self.assertFalse(os.path.exists(repo.worktreepath))

    def test_files_no_dir(self):
        """Test whether an empty list is returned when the input is none"""

        repo = GraalRepository('http://example.git', self.git_path)
        files = repo.files(None)
        self.assertEqual(files, [])

    def test_delete(self):
        """Test whether files and directories are deleted"""

        new_path = os.path.join(self.tmp_path, 'testworktree')

        repo = GraalRepository('http://example.git', self.git_path)
        repo.worktree(new_path)

        target_file = os.path.join(new_path, 'perceval/_version.py')
        target_folder = os.path.join(new_path, 'perceval/backends')

        self.assertTrue(os.path.exists(target_file))
        repo.delete(target_file)
        self.assertFalse(os.path.exists(target_file))

        self.assertTrue(os.path.exists(target_folder))
        repo.delete(target_folder)
        self.assertFalse(os.path.exists(target_folder))

        repo.prune()
        self.assertFalse(os.path.exists(repo.worktreepath))

    @staticmethod
    def __git_show_hash(repo):
        cmd_show = [GIT_EXEC_PATH, 'show']
        try:
            outs = repo._exec(cmd_show, cwd=repo.worktreepath, env=repo.gitenv)
        except Exception:
            return None

        outs = outs.decode("utf-8")
        hash = outs.split("\n")[0].replace("commit ", "")
        return hash


class TestGraalCommand(unittest.TestCase):
    """GraalCommand tests"""

    def setUp(self):
        super().setUp()
        self.tmp_path = tempfile.mkdtemp(prefix='graal_')

    def tearDown(self):
        super().tearDown()
        shutil.rmtree(self.tmp_path)

    def test_backend_class(self):
        """Test if the backend class is Graal"""

        self.assertIs(GraalCommand.BACKEND, Graal)

    @unittest.mock.patch('os.path.expanduser')
    def test_gitpath_init(self, mock_expanduser):
        """Test gitpath initialization"""

        mock_expanduser.return_value = os.path.join(self.tmp_path, 'testpath')
        args = ['http://example.com/']

        cmd = MockedGraalCommand(*args)
        self.assertEqual(cmd.parsed_args.git_path,
                         os.path.join(self.tmp_path, 'testpath/http://example.com/' + '-git'))

        args = ['http://example.com/',
                '--git-path', '/tmp/gitpath']

        cmd = MockedGraalCommand(*args)
        self.assertEqual(cmd.parsed_args.git_path, '/tmp/gitpath')

    def test_setup_cmd_parser(self):
        """Test if it parser object is correctly initialized"""

        parser = GraalCommand.setup_cmd_parser(Graal)
        self.assertIsInstance(parser, GraalCommandArgumentParser)

        args = ['http://example.com/',
                '--git-path', '/tmp/gitpath',
                '--tag', 'test']

        parsed_args = parser.parse(*args)
        self.assertEqual(parsed_args.uri, 'http://example.com/')
        self.assertEqual(parsed_args.git_path, '/tmp/gitpath')
        self.assertEqual(parsed_args.tag, 'test')
        self.assertEqual(parsed_args.from_date, DEFAULT_DATETIME)
        self.assertEqual(parsed_args.to_date, None)
        self.assertEqual(parsed_args.branches, None)
        self.assertFalse(parsed_args.latest_items)
        self.assertEqual(parsed_args.worktreepath, DEFAULT_WORKTREE_PATH)
        self.assertEqual(parsed_args.in_paths, None)
        self.assertEqual(parsed_args.out_paths, None)
        self.assertEqual(parsed_args.entrypoint, None)
        self.assertFalse(parsed_args.details)
        self.assertEqual(parser._backend, Graal)

        args = ['http://example.com/',
                '--git-path', '/tmp/gitpath',
                '--tag', 'test',
                '--from-date', '1975-01-01',
                '--to-date', '2099-01-01',
                '--branches', 'master', 'testing',
                '--latest-items',
                '--worktree-path', '/tmp/custom-worktrees/',
                '--in-paths', '*.py', '*.java',
                '--out-paths', '*.c',
                '--entrypoint', 'module',
                '--details']

        parsed_args = parser.parse(*args)
        self.assertEqual(parsed_args.uri, 'http://example.com/')
        self.assertEqual(parsed_args.git_path, '/tmp/gitpath')
        self.assertEqual(parsed_args.tag, 'test')
        self.assertEqual(parsed_args.from_date, str_to_datetime('1975-01-01'))
        self.assertEqual(parsed_args.to_date, str_to_datetime('2099-01-01'))
        self.assertEqual(parsed_args.branches, ['master', 'testing'])
        self.assertTrue(parsed_args.latest_items)
        self.assertEqual(parsed_args.worktreepath, '/tmp/custom-worktrees/')
        self.assertEqual(parsed_args.in_paths, ['*.py', '*.java'])
        self.assertEqual(parsed_args.out_paths, ['*.c'])
        self.assertEqual(parsed_args.entrypoint, 'module')
        self.assertTrue(parsed_args.details)

        parser = GraalCommand.setup_cmd_parser(Graal)
        self.assertIsInstance(parser, GraalCommandArgumentParser)

        args = ['http://example.com/',
                '--git-path', '/tmp/gitpath',
                '--tag', 'test',
                '--exec-path', '/tmp/execpath']

        parsed_args = parser.parse(*args)
        self.assertEqual(parsed_args.uri, 'http://example.com/')
        self.assertEqual(parsed_args.git_path, '/tmp/gitpath')
        self.assertEqual(parsed_args.exec_path, '/tmp/execpath')
        self.assertEqual(parsed_args.tag, 'test')
        self.assertEqual(parsed_args.from_date, DEFAULT_DATETIME)
        self.assertEqual(parsed_args.to_date, None)
        self.assertEqual(parsed_args.branches, None)
        self.assertFalse(parsed_args.latest_items)
        self.assertEqual(parsed_args.worktreepath, DEFAULT_WORKTREE_PATH)
        self.assertEqual(parsed_args.in_paths, None)
        self.assertEqual(parsed_args.out_paths, None)
        self.assertEqual(parsed_args.entrypoint, None)
        self.assertFalse(parsed_args.details)
        self.assertEqual(parser._backend, Graal)


class TesGraalFunctions(unittest.TestCase):
    """Graal functions tests"""

    def test_find_backends(self):
        backends = graal.graal.find_backends(graal)[0]
        for b in backends.keys():
            self.assertTrue(issubclass(backends.get(b), Graal))


class TestFetch(TestCaseRepo):
    """Unit tests for fetch function"""

    def test_items(self):
        """Test whether a set of items is returned"""

        args = {
            'uri': 'http://example.com/',
            'gitpath': self.git_path,
            'tag': 'test'
        }

        items = graal.graal.fetch(CommandBackend, args, CATEGORY_MOCKED)
        items = [item for item in items]

        self.assertEqual(len(items), 6)
        for i in items:
            self.assertEqual(i['category'], CATEGORY_MOCKED)

    def test_items_multiple_branches(self):
        """Test whether the working tree is created from the first branch in `branches`"""

        args = {
            'uri': 'http://example.com/',
            'gitpath': self.git_path,
            'tag': 'test',
            'branches': ['master', 'v1', 'v2']
        }

        with self.assertLogs(logger, level='WARNING') as cm:
            items = graal.graal.fetch(CommandBackend, args, CATEGORY_MOCKED)
            items = [item for item in items]
            self.assertEqual(cm.output[0], 'WARNING:graal.graal:Only the branch master will be analyzed')

        self.assertEqual(len(items), 6)
        for i in items:
            self.assertEqual(i['category'], CATEGORY_MOCKED)

    def test_items_no_category(self):
        """Test whether a set of items is returned"""

        args = {
            'uri': 'http://example.com/',
            'gitpath': self.git_path,
            'tag': 'test'
        }

        items = graal.graal.fetch(CommandBackend, args, None)
        items = [item for item in items]

        self.assertEqual(len(items), 6)
        for i in items:
            self.assertEqual(i['category'], CATEGORY_MOCKED)

    def test_items_on_error(self):
        """Test whether an exception is thrown when fetching items"""

        args = {
            'uri': 'http://example.com/',
            'gitpath': self.git_path,
            'tag': 'test'
        }

        items = graal.graal.fetch(ErrorCommandBackend, args, CATEGORY_MOCKED)

        with self.assertRaises(Exception):
            _ = [item for item in items]


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['graal',
 'graal.backends',
 'graal.backends.core',
 'graal.backends.core.analyzers',
 'graal.bin',
 'tests',
 'tests.data']

package_data = \
{'': ['*']}

install_requires = \
['bandit>=1.4.0',
 'cloc>=0.2.5,<0.3.0',
 'execnet>=1.9.0,<2.0.0',
 'flake8>=3.7.7',
 'grimoirelab-toolkit>=0.3.0,<0.4.0',
 'lizard==1.16.6',
 'networkx>=2.1',
 'perceval>=0.18,<0.19',
 'pydot>=1.2.4',
 'pylint>=1.8.4']

entry_points = \
{'console_scripts': ['graal = graal.bin.graal:main']}

setup_kwargs = {
    'name': 'graal',
    'version': '0.3.0',
    'description': 'A generic source code analyzer',
    'long_description': '# Graal: a Generic Repository AnALyzer [![Build Status](https://github.com/chaoss/grimoirelab-graal/workflows/tests/badge.svg)](https://github.com/chaoss/grimoirelab-graal/actions?query=workflow:tests+branch:master+event:push) [![Coverage Status](https://coveralls.io/repos/github/chaoss/grimoirelab-graal/badge.svg?branch=master)](https://coveralls.io/github/chaoss/grimoirelab-graal?branch=master)\n\nGraal leverages on the Git backend of [Perceval](https://github.com/chaoss/grimoirelab-perceval) and enhances it to set up ad-hoc\nsource code analysis. Thus, it fetches the commits from a Git repository and provides a mechanism to plug third party tools/libraries focused on source code analysis.\n\n## How it works\nThe Perceval Git backend creates a local mirror of a Git repository (local or remote), fetches the metadata of commits in chronological order and returns them as a list of JSON documents\n(one per commit). Graal leverages on the incremental functionalities provided by the Git backend and enhances the logic to handle Git repositories by creating a\nworking tree to perform checkout operations (which are not possible on a Git mirror). Graal intercepts each JSON document and enables the user to perform the following steps:\n- **Filter.** The filtering is used to select or discard commits based on the information available in the JSON document and/or via the Graal parameters. For any selected commit,\nGraal executes a checkout on the working tree using the commit hash, thus setting the state of the working tree at that given revision.\n- **Analyze.** The analysis takes the JSON document and the current working tree and enables the user to set up ad-hoc source code analysis by plugging existing tools through system calls\nor their Python interfaces, when possible. The results of the analysis are parsed and manipulated by the user and then automatically embedded in the JSON document. It is worth\nnoting that in this step the user can rely on some predefined functionalities of Graal to deal with the repository snapshot (e.g., listing files, creating archives).\n- **Post-process.** In the final step, the inflated JSON document can be optionally processed to alter (e.g., renaming, removing) its attributes, thus granting the user complete control over the output of Graal executions.\n\nSeveral parameters (inherited from the [Git backend](http://perceval.readthedocs.io/en/latest/perceval.backends.core.html#module-perceval.backends.core.git)) are available to control the execution; for instance, **from_date** and **to_date**\nallow to select commits authored since and before a given date, **branches** allows to fetch commits only from specific branches,\nand **latest_items** returns only those commits which are new since the last fetch operation. Graal includes additional parameters to drive\nthe analysis to filter in/out files and directories in the repository (**in_paths** and **out_paths**), set the **entrypoint**\nand define the **details** level of the analysis (useful when analyzing large software projects).\n\n## Requirements\n- [lizard](https://github.com/terryyin/lizard)==1.16.6\n- [perceval](https://github.com/chaoss/grimoirelab-perceval)>=0.9.6\n- [pylint](https://github.com/PyCQA/pylint)>=1.8.4\n- [flake8](https://github.com/PyCQA/flake8)>=3.7.7\n- [networkx](https://github.com/networkx/networkx)>=2.1\n- [pydot](https://github.com/pydot/pydot)>=1.2.4\n- [bandit](https://github.com/PyCQA/bandit)>=1.4.0\n- [grimoirelab-toolkit](https://github.com/chaoss/grimoirelab-toolkit)>=0.1.4\n- [cloc](http://cloc.sourceforge.net/)\n- [scc](https://github.com/boyter/scc)\n- [nomos](https://github.com/fossology/fossology/tree/master/src/nomos) \n- [scancode](https://github.com/nexB/scancode-toolkit) \n- [github-linguist](https://github.com/github/linguist) <=7.15\n\n### How to install/create the executables:\n- **Cloc**\n  ```\n  $> sudo apt-get install cloc\n  ```\n\n- **SCC**\n\nA tool similar to cloc - for counting physical the lines of code, blank lines, comment lines, and physical lines of source code in many programming languages and COCOMO estimates written in pure Go.\n\n  ```\n  $> go get -u github.com/boyter/scc/\n  ```\n\n\n- **Nomos**\n\nMaybe you\'ll need to install some packages for compiling the tool.\nFor example, in Debian, likely you\'ll need:\n\n```\nsudo apt-get install pkg-config libglib2.0-dev libjson-c-dev libpq-dev\n```\n\n- For compiling the tool (`nomossa`):\n\n```\n$> git clone https://github.com/fossology/fossology\n$> cd <...>/fossology/src/nomos/agent\n$> make -f Makefile.sa FO_LDFLAGS="-lglib-2.0 -lpq  -lglib-2.0 -ljson-c -lpthread -lrt"\n```\n\n- **ScanCode**\n\n```\ngit clone https://github.com/nexB/scancode-toolkit.git\ncd scancode-toolkit\ngit checkout -b test_scancli 96069fd84066c97549d54f66bd2fe8c7813c6b52\n./scancode --help\n```\n\n   **Note**: We\'re now using a clone of scancode-toolkit instead of a release, as the latest release is of 15th February 2019 and the `scancli.py` script (required for execution of scancode_cli) was incroporated later i.e 5th March 2019 and there hasn\'t been a release since.\n\n- **ScanCode Cli**\n\nAfter successfully executing the above mentioned steps, (if required) we have to install python modules: `simplejson` and `execnet`, for the execution of `scancode_cli` analyzer.\n\n```\npip install simplejson execnet\n```\n\n\n##  How to install/uninstall\nGraal is being developed and tested mainly on GNU/Linux platforms. Thus it is very likely it will work out of the box\non any Linux-like (or Unix-like) platform, upon providing the right version of Python (3.5, 3.6).\n\n\n**To install**, run:\n```\n$> git clone https://github.com/chaoss/grimoirelab-graal.git\n$> python3 setup.py build\n$> python3 setup.py install\n```\n\n**To uninstall**, run:\n```\n$> pip3 uninstall graal\n```\n\n## Backends\nSeveral backends have been developed to assess the genericity of Graal. Those backends leverage on source code analysis\ntools, where executions are triggered via system calls or their Python interfaces. In the current status, the backends\nmostly target Python code, however other backends can be easily developed to cover other programming languages. The\ncurrently available backends are:\n- **CoCom** gathers data about code complexity (e.g., cyclomatic complexity, LOC) from projects written in popular programming languages such as: C/C++, Java, Scala, JavaScript, Ruby, Python, Lua and Golang. It leverages on [Cloc](http://cloc.sourceforge.net/), [Lizard](https://github.com/terryyin/lizard) and [scc](https://github.com/boyter/scc). The tool can be exectued at `file` and `repository` levels activated with the help of category: `code_complexity_lizard_file` or `code_complexity_lizard_repository`.\n- **CoDep** extracts package and class dependencies of a Python module and serialized them as JSON structures, composed of edges and nodes, thus easing the bridging with front-end technologies for graph visualizations. It combines [PyReverse](https://pypi.org/project/pyreverse/) and [NetworkX](https://networkx.github.io/).\n- **CoQua** retrieves code quality insights, such as checks about line-code’s length, well-formed variable names, unused imported modules and code clones. It uses [PyLint](https://www.pylint.org/) and [Flake8](http://flake8.pycqa.org/en/latest/index.html). The tools can be activated by passing the corresponding category: `code_quality_pylint` or `code_quality_flake8`.\n- **CoVuln** scans the code to identify security vulnerabilities such as potential SQL and Shell injections, hard-coded passwords and weak cryptographic key size. It relies on [Bandit](https://github.com/PyCQA/bandit).\n- **CoLic** scans the code to extract license & copyright information. It currently supports [Nomos](https://github.com/fossology/fossology/tree/master/src/nomos) and [ScanCode](https://github.com/nexB/scancode-toolkit). They can be activated by passing the corresponding category: `code_license_nomos`, `code_license_scancode`, or `code_license_scancode_cli`.\n- **CoLang** gathers insights about code language distribution of a git repository. It relies on [Linguist](https://github.com/github/linguist) and [Cloc](http://cloc.sourceforge.net/) tools. They can be activated by passing the corresponding category: `code_language_linguist` or `code_language_cloc`.\n\n### How to develop a backend\nCreating your own backend is pretty easy, you only need to redefine the following methods of Graal:\n- **_filter_commit.** This method is used to select or discard commits based on the information available in the JSON document\nand/or via the Graal parameters (e.g., the commits authored by a given user or targeting a given software component).\nFor any selected commit, Graal executes a checkout on the working tree using the commit hash, thus setting the state of\nthe working tree at that given revision.\n- **_analyze.** This method takes the document and the current working tree and allows to connect existing tools through system calls or their Python interfaces, when possible.\nThe results of the analysis, parsed and manipulated by the user, are automatically embedded in the JSON document.\n- **_post.** This method allows to alter (e.g., renaming, removing) the attributes of the inflated JSON documents.\n\n## How to use\n\nGraal can be used from command line or directly from Python, both usages are described below.\n\n### From command line\nLaunching Graal from command line does not require much effort, but only some basic knowledge of GNU/Linux shell commands.\n\nThe example below shows how easy it is to fetch code complexity information from a Git repository. The **CoCom** backend\nrequires the URL where the repository is located (_https://github.com/chaoss/grimoirelab-perceval_) and the local path\nwhere to mirror the repository (_/tmp/graal-cocom_). Then, the JSON documents produced are redirected to the file _graal-cocom.test_.\n\n- **CoCom Backend**\n\n```\n$ graal cocom https://github.com/chaoss/grimoirelab-perceval --git-path /tmp/graal-cocom > /graal-cocom.test\nStarting the quest for the Graal.\nGit worktree /tmp/... created!\nFetching commits: ...\nGit worktree /tmp/... deleted!\nFetch process completed: .. commits inspected\nQuest completed.\n```\n\n- **CoLic Backend**\n\n```\ngraal colic https://github.com/chaoss/grimoirelab-toolkit --git-path /tmp/scancode_cli --exec-path /home/scancode-toolkit/etc/scripts/scancli.py --category code_license_scancode_cli\nStarting the quest for the Graal.\nGit worktree /tmp/... created!\nFetching commits: ...\nGit worktree /tmp/... deleted!\nFetch process completed: .. commits inspected\nQuest completed.\n```\n\nIn the above example, we\'re using scancode_cli analyzer. Similarly, we can use the scancode analyzer by providing the category as `code_license_scancode` and it\'s corresponding executable path.\n\n### From Python\nGraal’s functionalities can be embedded in Python scripts. Again, the effort of using Graal is minimum. In this case the user\nonly needs some knowledge of Python scripting. The example below shows how to use Graal in a script.\n\nThe **graal.backends.core.cocom** module is imported at the beginning of the file, then the **repo_uri** and **repo_dir** variables\nare set to the URI of the Git repository and the local path where to mirror it. These variables are used to initialize a\n**CoCom class object**. In the last line of the script, the commits inflated with the result of the analysis are retrieved\nusing the fetch method. The fetch method inherits its argument from **Perceval**, thus it optionally accept two Datetime\nobjects to gather only those commits after and before a given date, a list of branches to focus on specific development\nactivities, and a flag to collect the commits available after the last execution.\n\n```\n#! /usr/bin/env python3\nfrom graal.backends.core.cocom import CoCom\n\n# URL for the git repo to analyze\nrepo_uri = ’http://github.com/chaoss/grimoirelab-perceval’\n\n# directory where to mirror the repo\nrepo_dir = ’/tmp/graal-cocom’\n\n# Cocom object initialization\ncc = CoCom(uri=repo_uri, git_path=repo_dir)\n\n# fetch all commits\ncommits = [commit for commit in cc.fetch()]\n```\n\n## How to integrate it with Arthur\n[Arthur](https://github.com/chaoss/grimoirelab-kingarthur) is another tool of the [Grimoirelab ecosystem](https://chaoss.github.io/grimoirelab/). It was originally designed to allow to schedule\nand run Perceval executions at scale through distributed **Redis** queues, and store the obtained results in an **ElasticSearch** database.\n\nArthur has been extended to allow handling Graal tasks, which inherit from Perceval Git tasks. The code to make this extension possible is\navailable at: https://github.com/chaoss/grimoirelab-kingarthur/pull/33.\n\nInformation about Arthur is available at https://github.com/chaoss/grimoirelab-kingarthur.\n',
    'author': 'GrimoireLab Developers',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://chaoss.github.io/grimoirelab/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
